/*
    vision.h: contains the header files for pre-processing, features extraction, object-part composition and inference 
    Copyright (C) <2015>  <Safoura Rezapour Lakani>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
		
		In the case of questions, please contact "safoura.rezapour-lakani@uibk.ac.at"
*/
#ifndef _VISION_H_
#define _VISION_H_

#define  ARMA_DONT_USE_WRAPPER
#include <fstream>
#include <string>
#include <math.h>
#include <float.h>
#include "boost/multi_array.hpp"
#include <armadillo>
#include <Python.h>
#include <tinyxml.h>

#include <pcl/registration/transformation_estimation.h>
#include <pcl/point_types.h>
#include <pcl/point_cloud.h>
#include <pcl/kdtree/kdtree.h>
#include <pcl/io/pcd_io.h>
#include <pcl/visualization/cloud_viewer.h>
#include <pcl/features/normal_3d.h>
#include <pcl/kdtree/kdtree.h>
#include <pcl/ModelCoefficients.h>
#include <pcl/sample_consensus/method_types.h>
#include <pcl/sample_consensus/model_types.h>
#include <pcl/segmentation/sac_segmentation.h>
#include <pcl/filters/extract_indices.h>
#include <pcl/io/vtk_io.h>
#include <pcl/io/ply_io.h>
#include <pcl/PolygonMesh.h>
#include <pcl/io/vtk_lib_io.h>
#include <pcl/io/impl/vtk_lib_io.hpp>
#include <vtkSTLReader.h>
#include <vtkSTLWriter.h>
#include <pcl/surface/vtk_smoothing/vtk_utils.h>
#include <pcl/registration/warp_point_rigid.h>
#include <pcl/registration/icp.h>
#include <pcl/common/pca.h>
#include <pcl/filters/voxel_grid.h>
#include <pcl/registration/ia_ransac.h>
#include <pcl/segmentation/supervoxel_clustering.h>
#include <pcl/features/principal_curvatures.h>
#include <pcl/filters/passthrough.h>
#include <pcl/segmentation/region_growing.h>

typedef pcl::PointXYZ PointType;
typedef pcl::PointCloud<PointType> Cloud;
typedef typename Cloud::Ptr CloudPtr;
typedef typename Cloud::ConstPtr CloudConstPtr;

typedef pcl::PointXYZRGB ColorPointType;
typedef pcl::PointCloud<ColorPointType> ColorCloud;
typedef typename ColorCloud::Ptr ColorCloudPtr;
typedef typename ColorCloud::ConstPtr ColorCloudConstPtr;

using namespace pcl;
using namespace std;
using namespace arma;

enum Mode
{
  SAMPLE,
  LOAD,
  PARTS,
  PARTS_ALL,
  CLUSTER,
  EVALUATE,
	PART_SEG,
  ESTIMATE_REGION_MODELS,
	ESTIMATE_PATCH_THRESHOLD,
	COLLECT_STATISTICS,
	INVALID
};

struct execMode
{
	int dims;
  string patch_cls_path;
  Mode mode;
  string pro_path;
  string patch_cls_mat_path;
  string eval_path;
  string input_pcd;
  string input_dir;
	string patch_n_cls_mat;
};

struct model
{
  string id;
  PointCloud<pcl::Normal>::Ptr cloud_normal;
  CloudPtr cloud;
  map< uint32_t,Supervoxel<PointType>::Ptr > supervoxel_clusters;
  map<int,vector<int> > voxels_map;
  
  map<int,set<int> > segment_nns;
  map<int,set<int> > segments;
  map<int,CloudPtr> segment_voxels;
  map<int,PointCloud<pcl::Normal>::Ptr> segment_normals;
  map<int,pcl::Normal> segment_normal;

  map<int,set<int> > surfaces_nns;
  map<int,set<int> > surfaces;
  map<int,CloudPtr> surfaces_voxels;   
  map<int,PointCloud<pcl::Normal>::Ptr> surfaces_normals;
  map<int,pcl::PointCloud<pcl::PrincipalCurvatures>::Ptr> surface_cvts;

  map<int,vector<pcl::PointCloud<pcl::PrincipalCurvatures>::Ptr> > surface_voxel_cvts;
  map<int,vector<Cube<double> > > surface_desc;
 
  map<int,vector<Cube<double> > > surface_voxel_hist;
  map<int,vector<Cube<double> > > surface_patch_normal;
  map<int,vector<int> > surface_voxel_map;
  map<int,vector<CloudPtr> > surface_segment_pts;
  map<int,map<int,int> > surface_segment_map;
  map<int,vector<vector<PointType> > > axis_map;
  map<int,vector<Mat<double> > > trans_map;
  map<int,vector<Mat<double> > > center_map;
  map<int,vector<Cube<double> > > surface_patch_axis;

  map<int,set<int> > part_surface_map;
  map<int,Cube<double> > parts_hist;
  map<int,CloudPtr> parts; 
  map<int,set<int> > part_nns;
}; 

const bool save_ = false;
const int Bin_Angle = 8;
const double min_potential_ = 1e-3;
const double normal_rad_ = 0.01;
const bool agg_thresh_min_ = true;
const double prob_thresh_ = 0.5 + 1e-2;

struct Hcluster
{
  vector<Cube<double> > members;
  vector<int> members_ids;
  Cube<double> center;
  vector<CloudPtr> mems_pts;
	double dist_threshold;
};

class vision{
  public:
	bool dims_meter_;
  int model_size_;
  string file_name_;

  vision();
  ~vision();

  wchar_t *nstrws_convert(const char *raw) {
  	wchar_t *rtn = (wchar_t *) calloc(1, (sizeof(wchar_t) * (strlen(raw) + 1)));
    setlocale(LC_ALL,"en_US.UTF-8");
    mbstowcs(rtn, raw, strlen(raw));
    return rtn;
	}

	CloudPtr load_cloud(string file_name);
  PointCloud<pcl::Normal>::Ptr estimate_normal();
  vector<int> get_points_from_visualizer(const pcl::visualization::AreaPickingEvent &area);  
  void segment_supervoxel(model &cur_model);
  void superpixel_clustering(model &cur_model,double epsilon=0.1);


  model decompose_object(CloudPtr cloud);
  vector<string> get_all_files(string dir_name,string pattern=".pcd");

  double chi_hist_distance(Cube<double> hist1,Cube<double> hist2);
  double bha_hist_distance(Cube<double> hist1,Cube<double> hist2);
  double euc_hist_distance(Cube<double> hist1,Cube<double> hist2);
	double shape_context_distance(Cube<double> hist1,Cube<double> hist2);

  vector<Hcluster> cluster_features(vector<model> models);
  double agg_clustering(vector<model> models,vector<Hcluster> &clusters);
  double inter_hist_distance(Cube<double> hist1,Cube<double> hist2);
  double dist_cluster(Hcluster cluster1,Hcluster cluster2,Mat<double> dist_mat);

  pcl::PointCloud<pcl::PrincipalCurvatures>::Ptr estimate_curvature(PointCloud<pcl::Normal>::Ptr cloud_normal,CloudPtr cloud);
  vector<string> sample_data(vector<string> files,double percentage);
  double dist_cluster_avg(Hcluster cluster1,Hcluster cluster2);
  double dist_cluster_max(Hcluster cluster1,Hcluster cluster2);

  vector<model> read_prototypes(string path,bool is_prototype=true);
  vector<Hcluster> read_clusters(vector<string> dirs,vector<model> &models_all);
  void save_feature(model m);
  void make_code_book(vector<Hcluster> &clusters);

  vector<int> measure_similarity_codebook(vector<Hcluster> clusters,Cube<double> hist);
  void estimate_codebook_frequency(vector<Hcluster> clusters);
  vector<model> make_evaluation_models(string eval_path);

	void extract_feature_axis(model &m);
	void estimate_co_occurance(vector<model> models,vector<model> pro_models,vector<Hcluster> clusters);
	void estimate_n_co_occurance(vector<model> pro_models,vector<Hcluster> clusters);
	set<int> get_neighbor_parts(model m,vector<set<int> > parts,int ip);
	model refine_models(model m);
	vector<set<int> > merge_patches_to_regions(model m,vector<Hcluster> clusters);
	vector<set<int> > merge_regions_to_parts(model m,vector<Hcluster> clusters,vector<set<int> > regs);
	void cluster_surface_patch(model &cur_model);
	vector<Hcluster> read_clusters_patch(vector<string> paths,vector<model> &models_all);

	model read_model(vector<string> files,bool is_prototype=true);

	void extract_patch(CloudPtr cloud,string file_name);
	set<int> refine_patch_points( vector<PointType> pts,CloudPtr cloud,model &m,Mat<double> &dist,map<int,int> &sid_map,int c_id );
	ColorCloudPtr save_prototype(model m,set<int> surf_vec);
	vector<CloudPtr> save_segmented_part(model m,vector<set<int> > regs);
	double estimate_overlap(vector<CloudPtr> parts,model m);
	void visualize_surfaces(model m);
	vector<set<int> > refine_multiple_patches(vector<CloudPtr> clouds,CloudPtr cloud,model &m);
	void convert_to_csv(CloudPtr cloud,PointCloud<pcl::Normal>::Ptr cloud_normal);   

	void visualize_axis(CloudPtr cloud,vector<vector<PointType> > axis_list);    
	vector<vector<PointType> > get_axis(CloudPtr cloud,PointCloud<pcl::Normal>::Ptr cloud_normal);
	void init_axis();
    
	double dist_cluster_avg(Hcluster cluster1,Cube<double> desc);
	void make_parts_visualization(vector<vector<set<int> > > parts_h,vector<vector<set<int> > > parts_h_pts,model m);
	vector<set<int> > segment_parts(model m,vector<Hcluster> clusters);
	model refine_models(vector<model> models,vector<set<int> > &parts);
	
	vector<vector<int> > ids_;
	vector<model> models_;
	map<int,int> model_map;
	map<int,int> model_local_map;
	vector<Cube<double> > hists; 

	double threshold_;
	vector<double> pair_dists_;

	vector<double> pair_dists_patch_;

	int id_;
	map<string,vector<model> > model_prototypes_;  
	map<string,vector<double> > prototype_frequency_;
	map<string,vector<double> > prototype_occ_;

	map<string,double> prototype_rel_scale_;

	Mat<double> dist_prototype_;
	vector<model> eval_models;
	map<int,string> cluster_name_map_;
	double min_threshold_;

	Mat<double> occ_mat_;
	Mat<double> occ_mat_n_;
	Cube<double> occ_mat_tri_;
	map<int,Mat<double> > occ_mat_parts_;
	map<int,vec> occ_cls_parts_;
	map<int,Mat<double> > occ_mat_parts_cls_;
	Mat<double> patch_occ_n_;

	bool patch_;
	double min_angle_;
	Mat<double> dist_mat_;
	double potential_;
	PyObject *pName, *pModule;
	int parts_axis_call_;
	double ind_max;
	int id_cls_;
	int cls_max_;
	map<int,string> object_cls_;
	map<int,int> object_cls_map_;
	bool parts_cls_;
	bool object_ins_cls_;
	CloudPtr cloud_; 
	double min_dist_cls_;
	vector<vector<set<int> > > parts_h_;
	vector<vector<set<int> > > parts_h_pts_;
	map<int,int> region_id_;
	map<int,int> region_id_pre_;
	map<int,vector<int> > patch_cb_;
	map<int,vector<int> > part_cb_;
	execMode exec_mode_;
	map<string,Mode> mode_map_;
	bool agg_max_;
	string default_path_;
  double voxel_resolution;
  double seed_resolution;
	vector<double> dists_cb_;
	vector<double> dists_cb_all_;
	vector<double> regs_prob_;
	vector<int> ids_all_;	
	double region_threshold_;
	vector<Mat<double> > dist_mats_;
		
	map<int,vector<set<int> > > regs_all_;
	map<int,double> regs_potential_;
	vector<double> p_overlap_;
	vector<double> p_ids_;
	set<int> ids_seg_;

  map<int,int> model_id_;
  map<int,int> part_id_;
	bool debug_;
	double sample_perc_;
}; 
#endif

